import numpy as np

#klasa za generisanje statistike
#input - numpy array duzine koliko ima runova, svaki element numpy array je drugi numpy niz duzine koliko ima iteracija u runu
# a svaki element tog drugog niza je solution
#treba nam sledece
#index of best run i solutiona [i,j], kako bismo vratili najbolji solution
#niz sa najboljim vrednostima za svaki run
#statistika sa najboljim vrednostima za svaki run (best, mean, worst, std)
#broj selektovanih features za najbolje resenje
#dodatno, mozemo da trazimo najbolje resenje i statistiku za najbolje runove na osnovu fitnessa ili error-a, pa onda kao argument
#metode za inicijalizaciju uzimamo i objective=error (error), objective=objective (fitness, objective)
#dakle, sto se tice samih resenja, nas na kraju zanima samo najbolje resenje zbog vizualizacije
#svaki objekat ove klase predstavlja sve podatke o jednom algoritmu (najbolji runovi, statistika najboljih runova, najbolje resenje sa parametrima, itd.)
#obratiti paznju: u zavisnosti od vrednosti objective argumenta, imamo glavni indikator i pomocni (sekundarni) indikator
#na primer, ako je objective=objective, objective je glavni, a seknudarni indikator je error
#za oba indikatora se vadi statistika i radjeno je mapiranje da bi se na primer najbolji objective (u najboljem runu) podudario sa
#error-om u tom istom runu. Gde ako je glavni objective objective, error u tom runu gde je najbolji objective ne mora nuzno da bude najbolji,
#na ovo obratiti paznju

class Statistics:
    def __init__(self,dataset,no_classes,problem,objective,method, runtime, iterations,D,run_array,indicator='cappa'):

        #ovo je naziv indikatora, dakle imamo objective i jos jedan parametar koji je vazan, npr cohhen kappa score, sto je indikator
        #naravno, zavisi sta ce biti indikator
        self.indicator=indicator
        #naziv dataseta (koristimo za cuvanje fajlova,string)
        self.dataset = dataset
        #broj klasa u datasetu
        self.no_classes = no_classes
        #naziv problema, feature selection, hiperparam optimization, itd. (koristimo za cuvanje fajlova), FS - Feature selection, HT - hyper-parameters tuning, FH - feature selection + hyper-param tuning
        self.problem = problem
        #da li statistiku vadimo na osnovu error-a ili objective-a (fitness), objective="error" je error, a objective="objective" je objective
        #najbolji run za objective ne mora da bude nuzno i najbolji run za error, na ovo obratiti paznju, radimo i za jedno i za drugo!
        self.objective = objective
        #naziv metode, koristimo kasnije za postprocessing
        self.method = method
        #broj runova
        self.runtime = runtime
        #broj iteracija
        self.iterations = iterations
        #broj parametara za koje ima smisla cuvati vrednosti, npr. za FS to je koliko ima features, za ELM to je samo nn, itd
        self.D = D
        #glavni input niz (numpy array duzine koliko ima runova, svaki element numpy array je drugi numpy niz duzine koliko ima iteracija u runu
        # a svaki element tog drugog niza je solution, dakle cuvamo sva resenja u populaciji i onda gledamo statistiku i gledamo koje je najboje
        self.run_array = run_array
        #cuvamo sada podatke za indekse najboljeg resenja i niz sa najboljim vrednostima za svaki run, i je run, j je iteracija, cuvamo i convergence za najboljiu run
        #bez obzira da li radimo sa objective ili error, uvek cuvamo oba convergence grapha (i za error i za objective) da vidimo
        #koliko su jedan i drugi u najboljem runu za zadati kriterijum
        #vracamo niz najboljih runova i za objective i za error, gde je run_array_best za glavni indikator, a run_array_best1 za drugi indikator
        #npr. ako je objective=objective, run_array_best je za objective, a run_array_best1 je za error
        self.index_best_i, self.index_best_j,self.run_array_best,self.run_array_best1,self.best_run_convergence_objective,self.best_run_convergence_error = self.find_best()

        #dodajemo jos jedan isti element na kraj nizova konvergicje zbog generianja grafova konvergencije
        self.best_run_convergence_objective = np.insert(self.best_run_convergence_objective,len(self.best_run_convergence_objective),self.best_run_convergence_objective[-1])
        self.best_run_convergence_error = np.insert(self.best_run_convergence_error ,len(self.best_run_convergence_error),
                                                        self.best_run_convergence_error[-1])

        #sada na osnovu indeksa i i j najboljeg resenja cuvamo i samo najbolje resenje
        #najsigurnije po svim kriterijumima je da za najbolje resenje vracamo ono iz najboljeg runa i poslednje iteracije
        self.best_solution = self.run_array[self.index_best_i][-1]
        #self.best_solution = self.run_array[self.index_best_i][self.index_best_j]
        #sada cuvamo broj selektovanih featura za najbolje resenje
        self.best_feature_size = self.best_solution.feature_size
        #sada cuvamo parametre najboljeg resenja, nema smisla cuvati sve parametre, vec samo D koliko je dat kao argument
        self.best_params = self.best_solution.x[0:D]
        #sada cuvamo y_proba najboljeg resenja, to nam kasnije treba za vizualizaciju
        self.best_y_proba = self.best_solution.y_proba
        #cuvamo i y najboljeg resenja
        self.best_y = self.best_solution.y

        #problem sa array_best i array_best1 je taj sto nisu mapirani kada se racuna najbolji i najlosiji, itd.
        #treba mapirati drugi na prvi, pa da se prikazu analogne vrednosti


        #sada cuvamo statistiku za najbolje runove, best, worst, mean, std, itd za glavni objective
        self.best_run = np.min(self.run_array_best)
        #mapiramo drugi (best_run1) za glavni pomocu argmin i argmax funkcije za best i worst
        #za najbolji
        self.best_arg = np.argmin(self.run_array_best)
        #za najgori
        self.worst_arg = np.argmax(self.run_array_best)
        self.worst_run = np.max(self.run_array_best)
        self.mean_run = np.mean(self.run_array_best)
        self.median_run = np.median(self.run_array_best)
        self.std_run = np.std(self.run_array_best)
        self.var_run = np.var(self.run_array_best)

        # sada cuvamo statistiku za najbolje runove, best, worst, mean, std, itd za sporedni (drugi) objective
        #self.best_run1 = np.min(self.run_array_best1)
        #sada za sekundarni indikator, best_run1 koristimo index od onog gore best_run
        self.best_run1 = self.run_array_best1[self.best_arg]
        #isto kao gore za najgori koristimo argmax argument
        self.worst_run1 = self.run_array_best1[self.worst_arg]
        #self.worst_run1 = np.max(self.run_array_best1)
        self.mean_run1 = np.mean(self.run_array_best1)
        self.median_run1 = np.median(self.run_array_best1)
        self.std_run1 = np.std(self.run_array_best1)
        self.var_run1 = np.var(self.run_array_best1)

        # konacno, sada cuvamo i model najboljeg resenja
        self.best_model = self.best_solution.model

        # pravimo sada pomocne promenljive za diversity
        # diversity za objective i diversity za indikator, sta god da je

        self.diversity_objective = self.best_solution.diversity[0]  # ovo je objective
        self.diversity_indicator = self.best_solution.diversity[1]  # ovo je indikator




    #pomocne metode
    #metoda koja vraca vrednosti indeksa najboljeg resenja [i][j] na osnovu podataka iz run_array
    #ova metoda vraca indekse najboljeg resenja (ili za error ili objective), vraca listu najboljih runova i za objective i za error
    #glavni se cuva u "run_array_best" a drugi se cuva u "run_array_best1". Ako je objective="objective", glavni je objective, a ako je objective="error", glavni je error
    #run_array_best1 je uvek za ovaj drugi
    def find_best(self):
        #indeks najboljeg resenja u celom nizu runova
        run_index_smallest = 20
        #indeks najboljeg resenja u svakom pojedinacnom nizu sa iteracijama
        iter_index_smallest = 20
        # pomocni niz gde pamptimo najbolje vrednosti za svaki run za glavni indikator (zavisi da li je objective=objective ili objective=type)
        run_array_best = np.empty([self.runtime,],dtype=float)
        # pomocni niz gde pamptimo najbolje vrednosti za svaki run za drugi indikator (zavisi da li je objective=objective ili objective=type)
        run_array_best1 = np.empty([self.runtime, ], dtype=float)
        #pomocni niz gde pamtimo convergence najboljeg run-a za objective
        #bez obzira da li radimo za objective ili error (type), pravimo oba convergence speed grapha za najbolji run
        run_array_best_convergence_objective = np.empty([self.iterations,],dtype=float)
        # pomocni niz gde pamtimo convergence najboljeg run-a za error
        run_array_best_convergence_error = np.empty([self.iterations, ], dtype=float)

        # pomocna promenljiva da bismo nasli nabolju vrednost
        smallest = 30
        #proveravamo prvo da li radimo statistiku na osnovu error-a ili objectiva, type=0 je error, type=1 je objective
        if self.objective=="error":
            for i in range(len(self.run_array)):
                for j in range(len(self.run_array[i])):
                    if self.run_array[i][j].error < smallest:
                        smallest = self.run_array[i][j].error
                        run_index_smallest = i
                        iter_index_smallest = j
                # mozda error tokom iteracija i poraste, posto je objective primarni, pa je najsigurnije uzeti za best run array poslednju iteraciju
                run_array_best[i] = self.run_array[i][-1].error
                #run_array_best[i] = self.run_array[i][iter_index_smallest].error
                #najsigurnije je da ipak vracamo poslednji element niza, tj. poslednju iteraciju
                run_array_best1[i] = self.run_array[i][-1].objective_function
                #run_array_best1[i] = self.run_array[i][iter_index_smallest].objective_function
            # cuvamo najbolji convergence
            for k in range(len(self.run_array[run_index_smallest])):
                run_array_best_convergence_error[k] = self.run_array[run_index_smallest][k].error
                run_array_best_convergence_objective[k] = self.run_array[run_index_smallest][k].objective_function


        elif self.objective=="objective":
            for i in range(len(self.run_array)):
                for j in range(len(self.run_array[i])):
                    if self.run_array[i][j].objective_function < smallest:
                        smallest = self.run_array[i][j].objective_function
                        run_index_smallest = i
                        iter_index_smallest = j
                # najsigurnije je da ipak vracamo poslednji element niza, tj. poslednju iteraciju
                run_array_best[i] = self.run_array[i][-1].objective_function
                #run_array_best[i] = self.run_array[i][iter_index_smallest].objective_function
                #mozda error tokom iteracija i poraste, posto je objective primarni, pa je najsigurnije uzeti za best run array poslednju iteraciju
                run_array_best1[i] = self.run_array[i][-1].error
                #run_array_best1[i] = self.run_array[i][iter_index_smallest].error
            #cuvamo najbolji convergence
            for k in range(len(self.run_array[run_index_smallest])):
                run_array_best_convergence_error[k] = self.run_array[run_index_smallest][k].error
                run_array_best_convergence_objective[k] = self.run_array[run_index_smallest][k].objective_function
                #print(self.run_array[i][iter_index_smallest].objective_function)
        #vraca tuple sa indeksima najbolje resenja, i statistikom najboljih runova
        return (run_index_smallest, iter_index_smallest, run_array_best,run_array_best1,run_array_best_convergence_objective,run_array_best_convergence_error)


